;; Generated package description from engrave-faces.el  -*- no-byte-compile: t -*-
(define-package "engrave-faces" "0.3.0.0.20220511.140837" "Convert font-lock faces to other formats" '((emacs "27.1")) :commit "d6bc2228418941b9bee525cf588714569f6edb5d" :authors '(("TEC <https://github/tecosaur>")) :maintainer '("TEC" . "tec@tecosaur.com") :keywords '("faces") :url "https://github.com/tecosaur/engrave-faces")
